﻿using System;
using System.Threading;
using Microsoft.SPOT;
using Microsoft.SPOT.Hardware;
using Seletronica.NETMF.UPC;
using Seletronica.NETMF.UPC.Enums;
using WeatherStation.Sensors;

namespace WeatherStation
{
    public class Program
    {
        public static void Main()
        {
            // 1) Creo un bus I2C per comunicare con i sensori digitali
            SoftwareI2C bus = new SoftwareI2C(GPIO.B2, GPIO.A2, 100);
            // 2.1) Creo l'oggetto termometro/sensore di umidità collegato al bus I2C appena creato
            SHT1x thermometer = new SHT1x(bus, SHT1x.Voltage.Vdd_3_5V);
            // 2.2) Il barometro (anche lui sullo stesso bus digitale)
            MPL115A2 barometer = new MPL115A2(bus);
            // 3.1) L'anemometro lo collego all'opto isolatore A12
            Anemometer anemometer = new Anemometer((Cpu.Pin) Opto.OPTO_A12);
            // 3.2) mentre il pluviometro lo collego all' opto-isolatore B12
            Pluviometer pluviometer = new Pluviometer((Cpu.Pin)Opto.OPTO_B12);
            // 4) come ultimo sensore ho la direzione del vento che va collegata sull'ingresso analogico (B8 in questo esempio)
            WindDirectionUnit windDirectionUnit = new WindDirectionUnit(8);
            // 5) creo l'oggetto relè per la tenda parasole
            Relay awning = new Relay(Relays.RELAY_11);

            // Imposto come unità di misura della pressione i mBar
            barometer.MeasureUnit = MPL115A2.PressureMeasureUnit.mBar;
            
            // Ciclo principale dell'applicazione
            while (true)
            {
                // Ogni 10 secondi leggo i sensori, stampo i dati nella finestra di debug
                // Inoltre verifico che non ci sia forte vento o che piova, nel qual caso
                // Faccio un impulso sul relè (questo segnale serve nell'esempio specifico)
                // a segnalare alla centralina delle tende parasole di ritrarre la tenda
                Thread.Sleep(10000);
                try
                {
                    // Stampo nella finesta di debug i dati rilevati dai sensori
                    thermometer.Read();
                    Debug.Print("Temperatura: " + thermometer.Temperature + "°C");
                    Debug.Print("Umidità relativa: " + (int)thermometer.Humidity + "%");
                    barometer.Read();
                    Debug.Print("Pressione atmosferica: " + (int)barometer.Pressure + " mBar");
                    Debug.Print("Velocità del vento: " + anemometer.WindSpeed + " km/h");
                    windDirectionUnit.Read();
                    Debug.Print("Direzione del vento: " + windDirectionUnit.DirectionString);
                    Debug.Print("Precipitazioni: " + pluviometer.RainQuantity + " mm/h");

                    // Se il vento è più forte di 25 km/h oppure piove
                    if(anemometer.WindSpeed > 25 || pluviometer.RainQuantity > 0)
                    {
                        // ritraggo le tende parasole generando un impulso di un secondo sul relè
                        awning.On(1000);
                    }
                }
                catch (Exception ex)
                {
                    Debug.Print(ex.Message);
                }
                
            }
        }

    }
}
