using System;
using Microsoft.SPOT.Hardware;

namespace WeatherStation.Sensors
{
    /// <summary>
    /// Oggetto anemometro
    /// </summary>
    class Anemometer
    {
        /// <summary>
        /// Velocit del vento misurata
        /// </summary>
        private double _windSpeed;

        /// <summary>
        /// L'input a cui  collegata la sonda
        /// </summary>
        private readonly InterruptPort _port;

        /// <summary>
        /// data/ora ultimo campionamento
        /// </summary>
        private DateTime _lastPulseTime;

        /// <summary>
        /// Velocit del vento in km/h
        /// </summary>
        public double WindSpeed { 
            get
            {
                if(Utility.ElapsedMilliseconds(DateTime.Now, _lastPulseTime) > 5000)
                {
                    // Se sono passati pi di 5 secondi senza aver ricevuto un impulso considero la velocit del vento nulla
                    return 0.0;
                }
                return _windSpeed;
            }
        }

        /// <summary>
        /// Fattore di conversione da impulsi al secondo a km/h
        /// </summary>
        public double ConversionFactor { get; set; }

        /// <summary>
        /// Costruttore
        /// </summary>
        /// <param name="input"></param>
        public Anemometer(Cpu.Pin input)
        {
            _port = new InterruptPort(input, true, Port.ResistorMode.PullUp, Port.InterruptMode.InterruptEdgeHigh);
            _port.OnInterrupt += _port_OnInterrupt;

            // Dal sito http://www.philpot.me/weatherinsider.html si deduce che impulsi con frequenza 1Hz
            // equivalgono ad una velocit del vento di 2.4Km/h
            ConversionFactor = 2.4;
        }

        /// <summary>
        /// Evento richiamato quando lo switch reed commuta
        /// </summary>
        /// <param name="data1"></param>
        /// <param name="data2"></param>
        /// <param name="time"></param>
        void _port_OnInterrupt(uint data1, uint data2, DateTime time)
        {
            // sono passati x millisecondi dall'ultima transizione
            long period = Utility.ElapsedMilliseconds(time, _lastPulseTime);
            _windSpeed = (1000.0/period)*ConversionFactor;
            _lastPulseTime = time;
        }
    }
}
