using System.Threading;

namespace WeatherStation.Sensors
{

    /// <summary>
    /// Barometro digitale
    /// </summary>
    class MPL115A2
    {
        public enum PressureMeasureUnit
        {
            kPa,
            mBar
        }

        /// <summary>
        /// Il bus utilizzato per la comunicazione
        /// </summary>
        private readonly SoftwareI2C _bus;

        /// <summary>
        /// Coefficienti per la compensazione
        /// </summary>
        private double a0, b1, b2, c12;

        /// <summary>
        /// La pressione atmosferica in mbar
        /// </summary>
        public double Pressure { get; private set; }

        /// <summary>
        /// l'unit di misura in cui  espressa la misura della pressione
        /// </summary>
        public PressureMeasureUnit MeasureUnit { get; set; }

        /// <summary>
        /// Costruttore
        /// </summary>
        /// <param name="bus"></param>
        public MPL115A2(SoftwareI2C bus)
        {
            _bus = bus;
            MeasureUnit = PressureMeasureUnit.kPa;
        }

        /// <summary>
        /// Funzione per la conversione di un numero dalla notazione in virgola fissa a virgola mobile
        /// </summary>
        /// <param name="value">Il valore in virgola fissa da convertire</param>
        /// <param name="integerPartSize">il numero di bits che rappresentano la parte intera</param>
        /// <param name="fractionaPartSize">il numero di bits che rappresentano la parte frazionaria</param>
        /// <returns></returns>
        private static double FixedToDouble(int value, int integerPartSize, int fractionaPartSize)
        {
            int mask = 0;
            bool negativeNumber = false;
            if (fractionaPartSize > 0)
            {
                for (int i = 0; i < fractionaPartSize; i++)
                {
                    mask <<= 1;
                    mask++;
                }
            }

            if ((value & (1 << (1 + integerPartSize + fractionaPartSize))) != 0)
            {
                // numero negativo (in complemento a 2)
                value = ~value;
                value++;
                negativeNumber = true;
            }

            int intPart = value >> fractionaPartSize;
            int fracPart = value & mask;

            if (negativeNumber)
                return -(intPart + ((double)fracPart / (1 << fractionaPartSize)));
            return (intPart + ((double)fracPart / (1 << fractionaPartSize)));
        }

        /// <summary>
        /// Funzione per la lettura dei parametri di compensazione
        /// </summary>
        private void ReadCoefficients()
        {
            if (!_bus.Transmit(true, false, 0xC0)) return; // Indirizzo dello slave a 7bit + bit di scrittura
            if (!_bus.Transmit(false, false, 0x04)) return; // Comando di lettura coefficiente a 8 msb

            if (!_bus.Transmit(true, false, 0xC1)) return; // Indirizzo dello slave a 7bit + bit di lettura
            byte[] coeff = new byte[8];
            for (int i = 0; i < 8; i++)
            {
                coeff[i] = _bus.Receive(true, i == 7);
            }

            // Vedi nota applicativa AN3785 per il formato dei numeri
            short value = (short) ((coeff[0] << 8) + coeff[1]);
            a0 = FixedToDouble(value,12,3);

            value = (short)((coeff[2] << 8) + coeff[3]);
            b1 = FixedToDouble(value, 2, 13);

            value = (short)((coeff[4] << 8) + coeff[5]);
            b2 = FixedToDouble(value, 1, 14);

            value = (short)((coeff[6] << 8) + coeff[7]);
            c12 = FixedToDouble(value, 0, 24);
        }

        /// <summary>
        /// Lettura del sensore
        /// </summary>
        /// <returns></returns>
        public bool Read()
        {
            if(a0 == 0.0)
            {
                // La prima volta che viene eseguita la lettura vengono letti i coefficienti di compensazione dal sensore
                ReadCoefficients();
            }

            // Avvio la conversione
            if (!_bus.Transmit(true, false, 0xC0)) return false; // Indirizzo dello slave a 7bit + bit di scrittura
            if (!_bus.Transmit(false, false, 0x12)) return false; // Comando di inizio conversione (0x12)
            if (!_bus.Transmit(false, true, 0x00)) return false; // come descritto sul datasheet
            // Attendo 2 ms per dare il tempo al sensore di finire la conversione
            Thread.Sleep(2);
            // lettura della pressione e della temperatura 
            if (!_bus.Transmit(true, false, 0xC0)) return false; // Indirizzo dello slave a 7bit + bit di scrittura
            if (!_bus.Transmit(false, false, 0x00)) return false; // Comando di lettura della pressione MSB (0x00)
            if (!_bus.Transmit(true, false, 0xC1)) return false; // Indirizzo dello slave a 7bit + bit di lettura
            byte pressureMsb = _bus.Receive(true, false);
            byte pressureLsb = _bus.Receive(true, false);
            int pAdc = (pressureMsb << 8 + pressureLsb)>>6;
            byte temperatureMsb = _bus.Receive(true, false);
            byte temperatureLsb = _bus.Receive(true, true);
            int tAdc = (temperatureMsb << 8 + temperatureLsb) >> 6;

            // Applico le formule descritte nella nota applicativa AN3785 per calcolare la pressione atmosferica (compensata intemperatura)
            double pComp =  a0 + (b1 + c12 * tAdc) * pAdc + b2 * tAdc;

            // A questo punto calcolo la pressione in kPa sempre secondo le formule fornite dal costruttore
            double kPa = pComp*((115 - 50)/1023.0) + 50;

            if (MeasureUnit == PressureMeasureUnit.kPa)
                Pressure = kPa;
            else if (MeasureUnit == PressureMeasureUnit.mBar)
                Pressure = 10 * kPa;

            return true;
        }
    }
}
