using System;
using Microsoft.SPOT.Hardware;

namespace WeatherStation.Sensors
{
    /// <summary>
    /// Oggetto pluviometro
    /// </summary>
    class Pluviometer
    {
        /// <summary>
        /// Velocit del vento misurata
        /// </summary>
        private double _rainQuantity;

        /// <summary>
        /// L'input a cui  collegata la sonda
        /// </summary>
        private readonly InterruptPort _port;

        /// <summary>
        /// data/ora ultimo campionamento
        /// </summary>
        private DateTime _lastPulseTime;

        /// <summary>
        /// Quantit di pioggia in mm all'ora
        /// </summary>
        public double RainQuantity { 
            get
            {
                if(Utility.ElapsedSeconds(DateTime.Now, _lastPulseTime) > 2000)
                {
                    // Se sono passati pi di 2000 secondi (33 minuti circa) senza aver ricevuto un impulso considero la quantit di pioggia nulla
                    // in quanto 3600*0.2794/2000 porta a circa 0.5 mm/ora
                    return 0.0;
                }
                return _rainQuantity;
            }
        }

        /// <summary>
        /// Fattore di conversione (mm di pioggia per ogni impulso)
        /// </summary>
        public double ConversionFactor { get; set; }

        /// <summary>
        /// Costruttore
        /// </summary>
        /// <param name="input"></param>
        public Pluviometer(Cpu.Pin input)
        {
            _port = new InterruptPort(input, true, Port.ResistorMode.PullUp, Port.InterruptMode.InterruptEdgeHigh);
            _port.OnInterrupt += _port_OnInterrupt;

            // Dal sito http://www.philpot.me/weatherinsider.html si deduce che il sensore commuta quando scendono 0.011" di piggia
            // cio 0.2794 mm
            ConversionFactor = 0.2794;
        }

        /// <summary>
        /// Evento richiamato quando lo switch reed commuta
        /// </summary>
        /// <param name="data1"></param>
        /// <param name="data2"></param>
        /// <param name="time"></param>
        void _port_OnInterrupt(uint data1, uint data2, DateTime time)
        {
            // sono passati x secondi dall'ultima transizione
            double period = Utility.ElapsedSeconds(time, _lastPulseTime);
            // se per una transazione sono scesi ConversionFactor mm di pioggia
            // allora la quantit di pioggia in un secondo :
            double quantity = ConversionFactor / period; // in mm/s
            // dato che lo voglio esprimere in mm/h e 1 ora = 3600 s
            _rainQuantity = quantity * 3600;
            _lastPulseTime = time;
        }
    }
}
