﻿using System;
using System.Threading;

namespace WeatherStation.Sensors
{
    /// <summary>
    /// Driver per il sensore di temperatura/umidità SHT1X
    /// </summary>
    public class SHT1x
    {
        #region Costanti ...

        // Indirizzo;
        private const byte ADDR = 0x00;

        // Comandi (5 bits)
        private const byte CMD_MEASURE_TEMP = 0x03;     // b00011
        private const byte CMD_MEASURE_RH = 0x05;       // b00101
        private const byte CMD_READ_STATUS_REG = 0x07;  // b00111
        private const byte CMD_WRITE_STATUS_REG = 0x06; // b00110
        private const byte CMD_SOFT_RESET = 0x1E;       // b11110

        // d1, d2 coefficienti per la temperatura
        // Celsius
        private const double D1_VDD5V_C = -40.1F;
        private const double D1_VDD4V_C = -39.8F;
        private const double D1_VDD3_5V_C = -39.7F;
        private const double D1_VDD3V_C = -39.6F;
        private const double D1_VDD2_5V_C = -39.4F;
        private const double D2_14BIT_C = 0.01F;
        private const double D2_12BIT_C = 0.04F;
        // Farenheit
        private const double D1_VDD5V_F = -40.2F;
        private const double D1_VDD4V_F = -39.6F;
        private const double D1_VDD3_5V_F = -39.5F;
        private const double D1_VDD3V_F = -39.3F;
        private const double D1_VDD2_5V_F = -38.9F;
        private const double D2_14BIT_F = 0.018F;
        private const double D2_12BIT_F = 0.072F;

        // c1, c2, c3, t1, t2 coefficienti per l'umidità
        // 12 bit resolution
        private const double C1_12BIT = -2.0468F;
        private const double C2_12BIT = 0.0367F;
        private const double C3_12BIT = -1.5955E-6F;
        private const double T1_12BIT = 0.01F;
        private const double T2_12BIT = 0.00008F;
        // 8 bit resolution
        private const double C1_8BIT = -2.0468F;
        private const double C2_8BIT = 0.5872F;
        private const double C3_8BIT = -4.0845E-4F;
        private const double T1_8BIT = 0.01F;
        private const double T2_8BIT = 0.00128F;

        // maschere di bit per il registro di stato
        private const byte END_OF_BATTERY_MASK = 0x40;
        private const byte HEATER_MASK = 0x04;
        private const byte NO_RELOAD_FROM_OTP_MASK = 0x02;
        private const byte LESS_RESOLUTION = 0x01;

        #endregion

        #region Membri privati ...

        private readonly Voltage _voltage;
        private readonly SoftwareI2C _bus;

        #endregion

        /// <summary>
        /// Temperatura in gradi Celsius
        /// </summary>
        public double Temperature { get; private set; }

        /// <summary>
        /// Umidità relativa in %
        /// </summary>
        public double Humidity { get; private set; }

        /// <summary>
        /// Costruttore
        /// </summary>
        /// <param name="bus">Il bus I2C da utilizzare per la comunicazione </param>
        /// <param name="voltage">La tensione di alimentazione del sensore</param>
        public SHT1x(SoftwareI2C bus, Voltage voltage)
        {
            _bus = bus;
            _voltage = voltage;
        }

        /// <summary>
        /// Esegue la sequenza di start del sensore
        /// </summary>
        private void StartTransmission()
        {
            if (!_bus.SDA.Active)
                // imposto SDA come output digitale
                _bus.SDA.Active = true;

            if (!_bus.SCL.Active)
                // imposto SCL come output digitale
                _bus.SCL.Active = true;

            // sequenza di start (vedi datasheet)
            _bus.SCL.Write(true);
            _bus.SDA.Write(false);
            _bus.SCL.Write(false);
            _bus.SCL.Write(true);
            _bus.SDA.Write(true);
            _bus.SCL.Write(false);
        }

        /// <summary>
        /// Funzione per spedire un comando al sensore
        /// </summary>
        /// <param name="cmd">Il comando da spedire</param>
        /// <returns>ACK recevuto dal sensore</returns>
        private bool SendCommand(byte cmd)
        {
            byte send = (byte)(ADDR | cmd);

            if (!_bus.SDA.Active)
                _bus.SDA.Active = true;

            // sequenza di start (vedi datasheet)
            StartTransmission();

            return WriteByte(send);
        }

        /// <summary>
        /// Lettura di un byte
        /// </summary>
        /// <param name="sendAck">Indica se spedire o meno l'ACK</param>
        /// <returns>Il byte letto dal sensore</returns>
        private byte ReadByte(bool sendAck)
        {
            byte resp = 0x00;

            if (_bus.SDA.Active)
                _bus.SDA.Active = false;

            // lettura dei bit a partire dall'MSB
            for (int i = 7; i >= 0; i--)
            {
                _bus.SCL.Write(true);
                resp |= (_bus.SDA.Read()) ? (byte)(1 << i) : (byte)0;
                _bus.SCL.Write(false);
            }

            _bus.SDA.Active = true;
            // spedisce ack se richiesto
            _bus.SDA.Write(!sendAck);
            _bus.SCL.Write(true);
            _bus.SCL.Write(false);
            _bus.SDA.Write(true);

            return resp;
        }

        /// <summary>
        ///Invio di un byte
        /// </summary>
        /// <param name="data">Il byte da spedire</param>
        /// <returns>ACK dal sensore</returns>
        private bool WriteByte(byte data)
        {
            if (!_bus.SDA.Active)
                _bus.SDA.Active = true;

            for (int i = 7; i >= 0; i--)
            {
                byte b = (byte)(data & (1 << i));

                _bus.SDA.Write(b != 0);
                _bus.SCL.Write(true);
                _bus.SCL.Write(false);
            }

            _bus.SDA.Active = false;
            _bus.SCL.Write(true);
            bool ack = _bus.SDA.Read();
            _bus.SCL.Write(false);

            return ack;
        }

        /// <summary>
        /// Lettura della temperatura
        /// </summary>
        /// <returns></returns>
        private int ReadTemperatureRaw()
        {
            SendCommand(CMD_MEASURE_TEMP);

            if (_bus.SDA.Active)
                _bus.SDA.Active = false;

            // aspetto LOW su SDA (fine conversione)
            DateTime start = DateTime.Now;
            while (_bus.SDA.Read())
            {
                Thread.Sleep(10);
                if (Utility.ElapsedMilliseconds(DateTime.Now, start) > 500)
                {
                    return 0;
                }
            }

            int resp = 0;
            resp |= ReadByte(true);

            resp = resp << 8;
            resp |= ReadByte(false);

            return resp;
        }

        /// <summary>
        /// Lettura umidità
        /// </summary>
        /// <returns></returns>
        private int ReadRelativeHumidityRaw()
        {
            SendCommand(CMD_MEASURE_RH);

            if (_bus.SDA.Active)
                _bus.SDA.Active = false;

            // aspetto LOW su SDA (fine conversione)
            DateTime start = DateTime.Now;
            while (_bus.SDA.Read())
            {
                Thread.Sleep(10);
                if(Utility.ElapsedMilliseconds(DateTime.Now, start)>500)
                {
                    return 0;
                }
            }

            int resp = 0;
            resp |= ReadByte(true);

            resp = resp << 8;
            resp |= ReadByte(false);

            return resp;
        }

        /// <summary>
        /// Reset Connection
        /// </summary>
        public void ResetConnection()
        {
            if (!_bus.SDA.Active)
                _bus.SDA.Active = true;

            _bus.SDA.Write(true);
            // toggle 9 volte del segnale di il clock (vedi datasheet)
            for (int i = 0; i < 10; i++)
            {
                _bus.SCL.Write(true);
                _bus.SCL.Write(false);
            }
        }

        /// <summary>
        /// Lettura registro di stato
        /// </summary>
        /// <returns></returns>
        private byte ReadStatusRegister()
        {
            Thread.Sleep(10);
            SendCommand(CMD_READ_STATUS_REG);
            byte status = ReadByte(false);
            return status;
        }

        /// <summary>
        /// Scrittura registro di stato
        /// </summary>
        /// <param name="status">il valore da scrivere</param>
        /// <returns>true se ricevo Ack in risposta</returns>
        private bool WriteStatusRegister(byte status)
        {
            SendCommand(CMD_WRITE_STATUS_REG);
            return WriteByte(status);
        }

        /// <summary>
        /// Lettura temperatura
        /// </summary>
        /// <param name="unit">unità di Temperatura (°C or °F)</param>
        /// <returns>il valore di temperatura</returns>
        private double ReadTemperature(TempUnit unit)
        {
            double d1 = 0, d2 = 0;
            // T = d1 + d2 * SOt

            // determino d1 in base alla temperatura e alla tensione di alimentazione
            if (unit == TempUnit.Celsius)
            {
                switch (_voltage)
                {
                    case Voltage.Vdd_5V:
                        d1 = D1_VDD5V_C;
                        break;
                    case Voltage.Vdd_4V:
                        d1 = D1_VDD4V_C;
                        break;
                    case Voltage.Vdd_3_5V:
                        d1 = D1_VDD3_5V_C;
                        break;
                    case Voltage.Vdd_3V:
                        d1 = D1_VDD3V_C;
                        break;
                    case Voltage.Vdd_2_5V:
                        d1 = D1_VDD2_5V_C;
                        break;
                }
            }
            else if (unit == TempUnit.Farenheit)
            {
                switch (_voltage)
                {
                    case Voltage.Vdd_5V:
                        d1 = D1_VDD5V_F;
                        break;
                    case Voltage.Vdd_4V:
                        d1 = D1_VDD4V_F;
                        break;
                    case Voltage.Vdd_3_5V:
                        d1 = D1_VDD3_5V_F;
                        break;
                    case Voltage.Vdd_3V:
                        d1 = D1_VDD3V_F;
                        break;
                    case Voltage.Vdd_2_5V:
                        d1 = D1_VDD2_5V_F;
                        break;
                }
            }

            // determino d2, in base all'unità di temperatura e i bit di risoluzione
            byte status = ReadStatusRegister();

            // less resolution : 8bit RH, 12bit Temp
            if ((status & LESS_RESOLUTION) != 0x00)
            {
                if (unit == TempUnit.Celsius)
                    d2 = D2_12BIT_C;
                else if (unit == TempUnit.Farenheit)
                    d2 = D2_12BIT_F;
            }
            // high resolution : 12bit RH, 14bit Temp
            else
            {
                if (unit == TempUnit.Celsius)
                    d2 = D2_14BIT_C;
                else if (unit == TempUnit.Farenheit)
                    d2 = D2_14BIT_F;
            }

            double temp = d1 + d2 * ReadTemperatureRaw();

            return temp;
        }

        /// <summary>
        /// Lettura del sensore
        /// </summary>
        public void Read()
        {
            double c1, c2, c3;
            double t1, t2;

            ReadStatusRegister();
            byte status = ReadStatusRegister();

            // 8 bit
            if ((status & LESS_RESOLUTION) != 0x00)
            {
                c1 = C1_8BIT;
                c2 = C2_8BIT;
                c3 = C3_8BIT;
                t1 = T1_8BIT;
                t2 = T2_8BIT;
            }
            // 12 bit
            else
            {
                c1 = C1_12BIT;
                c2 = C2_12BIT;
                c3 = C3_12BIT;
                t1 = T1_12BIT;
                t2 = T2_12BIT;
            }

            double SOrh = ReadRelativeHumidityRaw();

            double RHlinear = c1 + (c2 * SOrh) + (c3 * (SOrh * SOrh));

            // Leggo la temperatura per compensare il dato
            double temp = ReadTemperature(TempUnit.Celsius);

            double RH = ((temp - 25.0F) * (t1 + (t2 * SOrh))) + RHlinear;

            Humidity = RH;
            Temperature = temp;
        }

        /// <summary>
        /// soft reset
        /// </summary>
        /// <returns>ACK ricevuto in risposta al comando</returns>
        public bool SoftReset()
        {
            bool ack = SendCommand(CMD_SOFT_RESET);

            // occorre attendere almeno 11 ms (datasheet)
            Thread.Sleep(20);

            return ack;
        }

        /// <summary>
        /// Unità di temperatura
        /// </summary>
        public enum TempUnit
        {
            Celsius,
            Farenheit
        }

        /// <summary>
        /// Tensione di alimentazione
        /// </summary>
        public enum Voltage
        {
            Vdd_5V,
            Vdd_4V,
            Vdd_3_5V,
            Vdd_3V,
            Vdd_2_5V
        }
    }
}
