using System;
using Seletronica.NETMF.UPC;
using Seletronica.NETMF.UPC.Enums;

namespace WeatherStation.Sensors
{
    public enum WindDirection
    {
        N,
        NNE,
        NE,
        ENE,
        E,
        ESE,
        SE,
        SSE,
        S,
        SSO,
        SO,
        OSO,
        O,
        ONO,
        NO,
        NNO
    }

    class WindDirectionUnit
    {
        private readonly int[] _valueMap = new [] {33000,6570,8200,891,1000,678,2100,1410,3900,3130,16000,14120,120000,42120,64900,21880};

        /// <summary>
        /// L'ingresso analogico utilizzato per acquisire i dati
        /// </summary>
        private readonly AnalogIn _ain;
        
        /// <summary>
        /// Resistenza in Ohm della parte alta del partitore di tensione (vedi schematico dello schedino Dual ADC Module canale B)
        /// </summary>
        private const int Rb = 1000;
        
        /// <summary>
        /// La tensione di riferimento dell'A/D
        /// </summary>
        private const double Vref = 2.5;

        /// <summary>
        /// La direzione del vento
        /// </summary>
        public WindDirection Direction { get; private set; }

        /// <summary>
        /// La direzione del vento espressa come stringa
        /// </summary>
        public string DirectionString
        {
            get
            {
                switch (Direction)
                {
                    case WindDirection.N:
                        return "N";
                    case WindDirection.NNE:
                        return "NNE";
                    case WindDirection.NE:
                        return "NE";
                    case WindDirection.ENE:
                        return "ENE";
                    case WindDirection.E:
                        return "E";
                    case WindDirection.ESE:
                        return "ESE";
                    case WindDirection.SE:
                        return "SE";
                    case WindDirection.SSE:
                        return "SSE";
                    case WindDirection.S:
                        return "S";
                    case WindDirection.SSO:
                        return "SSO";
                    case WindDirection.SO:
                        return "SO";
                    case WindDirection.OSO:
                        return "OSO";
                    case WindDirection.O:
                        return "O";
                    case WindDirection.ONO:
                        return "ONO";
                    case WindDirection.NO:
                        return "NO";
                    case WindDirection.NNO:
                        return "NNO";
                    default:
                        return "Error";
                }
            }
        }
        /// <summary>
        /// Costruttore
        /// </summary>
        /// <param name="module">Il modulo su cui  connesso il sensore</param>
        /// <remarks>Valori validi del parametro module sono 3,4,6,7 e 8. Ogni altro valore genera un'eccezione (vedi implementazione)</remarks>
        public WindDirectionUnit(int module)
        {
            ADC channel;

            switch (module)
            {
                case 3:
                    channel = ADC.ADC_B3;
                    break;
                case 4:
                    channel = ADC.ADC_B4;
                    break;
                case 6:
                    channel = ADC.ADC_B6;
                    break;
                case 7:
                    channel = ADC.ADC_B7;
                    break;
                case 8:
                    channel = ADC.ADC_B8;
                    break;
                default:
                    throw new ArgumentException("Lo slot " + module + "non supporta la funzionalit A/D con lettura di resistenza");
            }

            // Imposto un canale analogico 0-2.5V (VREF=2.5V)
            _ain = new AnalogIn(channel, 2.5, 0);
        }

        /// <summary>
        /// Funzione per la lettura del sensore
        /// </summary>
        public void Read()
        {
            double volt = _ain.Read();
            // Dalla lettura in volt del convertitore e la funzione inversa del partitore di tensione, calcolo la resistenza (in Ohm)
            double r = Rb * volt / (Vref - volt);

            for (int i = 0; i < 16; i++)
            {
                // verifico che il valore di resistenza letto sia +/-10% del valore di riferimento (_valueMap)
                // questi valori sono stati rilevati sperimentalemente. per maggiori dettagli si rimanda al sito internet
                // http://www.philpot.me/weatherinsider.html
                if(_valueMap[i]*0.9 < r && r < _valueMap[i]*1.1)
                {
                    Direction = (WindDirection) i;
                    return;
                }
            }

        }
    }
}
